@@ -1,56 +1,52 @@
-/*
- *   bin86/ld/writebin.c
- *
- *   Copyright (C) 1992 Bruce Evans
- */
-
-extern long text_base_address;
-#define btextoffset text_base_address
-static long bdataoffset;
-#define page_size() 4096
 
 /* writebin.c - write binary file for linker */
 
-#ifdef A_OUT_H
-# include A_OUT_H
-#else
-# ifdef BSD_A_OUT
-#  include "bsd-a.out.h"
-#  define A_MINHDR 0
-#  define C_EXT N_EXT
-#  define C_STAT 0
-#  define n_was_name n_un.n_name
-#  define n_was_numaux n_other
-#  define n_was_other n_numaux
-#  define n_was_sclass n_type
-#  define n_was_strx n_un.n_strx
-#  define n_was_type n_desc
-# else /* not BSD_A_OUT */
-#  include "a.out.h"		/* maybe local copy of <a.out.h> for X-link */
-#  define n_was_name n_name
-#  define n_was_numaux n_numaux
-#  define n_was_other n_other
-#  define n_was_sclass n_sclass
-#  define n_was_strx n_value
-#  define n_was_type n_type
-# endif /* BSD_A_OUT */
+/* Copyright (C) 1994 Bruce Evans */
+
+#ifndef NO_AOUT
+#ifndef A_OUT_INCL
+#define A_OUT_INCL <a.out.h>
+#endif
 #endif
 
+#include "syshead.h"
+#include "bindef.h"
 #include "const.h"
 #include "obj.h"
 #include "type.h"
-#undef EXTERN
 #include "globvar.h"
 
+#ifdef AOUT_DETECTED
+#define btextoffset (text_base_value)
+#define bdataoffset (data_base_value)
+#define page_size() 4096
+
+#ifdef __ELF__
+#ifndef ELF_SYMS
+#define ELF_SYMS 1
+#endif
+#endif
+
 #ifdef EDOS
 # define FILEHEADERLENGTH 0
 #endif
 #ifdef MINIX
 # ifdef BSD_A_OUT
-#  define FILEHEADERLENGTH sizeof (struct exec)
+#  ifdef STANDARD_GNU_A_OUT
+#   define HEADERLEN (sizeof(struct exec))
+#  else
+#   define HEADERLEN (48)
+#  endif
 # else
-#  define FILEHEADERLENGTH A_MINHDR
+#  ifdef REL_OUTPUT
+#   define HEADERLEN (reloc_output?sizeof(struct exec):A_MINHDR)
 				/* part of header not counted in offsets */
+#  else
+#   define HEADERLEN (A_MINHDR)
+#  endif
+# endif
+# ifndef FILEHEADERLENGTH
+#  define FILEHEADERLENGTH (headerless?0:HEADERLEN)
 # endif
 #endif
 #define DPSEG 2
@@ -80,33 +76,33 @@ static long bdataoffset;
 #define memsizeof(struc, mem) sizeof(((struc *) 0)->mem)
 
 PRIVATE bool_t bits32;		/* nonzero for 32-bit executable */
-PRIVATE offset_t combase[NSEG];	/* bases of common parts of segments */
-PRIVATE offset_t comsz[NSEG];	/* sizes of common parts of segments */
-PRIVATE char curseg;		/* current segment, 0 to $F */
-PRIVATE offset_t edataoffset;	/* end of data */
-PRIVATE offset_t endoffset;	/* end of bss */
-PRIVATE offset_t etextoffset;	/* end of text */
-PRIVATE offset_t etextpadoff;	/* end of padded text */
-#ifdef BSD_A_OUT
+PRIVATE bin_off_t combase[NSEG];/* bases of common parts of segments */
+PRIVATE bin_off_t comsz[NSEG];	/* sizes of common parts of segments */
+PRIVATE fastin_t curseg;	/* current segment, 0 to $F */
+PRIVATE bin_off_t edataoffset;	/* end of data */
+PRIVATE bin_off_t endoffset;	/* end of bss */
+PRIVATE bin_off_t etextoffset;	/* end of text */
+PRIVATE bin_off_t etextpadoff;	/* end of padded text */
+#ifdef REL_OUTPUT
 PRIVATE unsigned ndreloc;	/* number of data relocations */
 #endif
 PRIVATE unsigned nsym;		/* number of symbols written */
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 PRIVATE unsigned ntreloc;	/* number of text relocations */
-PRIVATE bool_t reloc_output;	/* nonzero to leave reloc info in output */
+extern bool_t reloc_output;	/* nonzero to leave reloc info in output */
 #endif
 PRIVATE unsigned relocsize;	/* current relocation size 1, 2 or 4 */
-PRIVATE offset_t segadj[NSEG];	/* adjusts (file offset - seg offset) */
+PRIVATE bin_off_t segadj[NSEG];	/* adjusts (file offset - seg offset) */
 				/* depends on zero init */
-PRIVATE offset_t segbase[NSEG];	/* bases of data parts of segments */
+PRIVATE bin_off_t segbase[NSEG];/* bases of data parts of segments */
 PRIVATE char segboundary[9] = "__seg0DH";
 				/* name of seg boundary __seg0DL to __segfCH */
-PRIVATE offset_t segpos[NSEG];	/* segment positions for current module */
-PRIVATE offset_t segsz[NSEG];	/* sizes of data parts of segments */
+PRIVATE bin_off_t segpos[NSEG];	/* segment positions for current module */
+PRIVATE bin_off_t segsz[NSEG];	/* sizes of data parts of segments */
 				/* depends on zero init */
 PRIVATE bool_t sepid;		/* nonzero for separate I & D */
 PRIVATE bool_t stripflag;	/* nonzero to strip symbols */
-PRIVATE offset_t spos;		/* position in current seg */
+PRIVATE bin_off_t spos;		/* position in current seg */
 PRIVATE bool_t uzp;		/* nonzero for unmapped zero page */
 
 #ifdef EDOS
@@ -114,79 +110,24 @@ FORWARD unsigned binheaderlength P((char *commandname));
 FORWARD char *idconvert P((struct entrylist *elptr, char *commandname));
 #endif
 FORWARD void linkmod P((struct modstruct *modptr));
-FORWARD void linkrefs P((struct modstruct *modptr));
 FORWARD void padmod P((struct modstruct *modptr));
-FORWARD void setsym P((char *name, offset_t value));
+FORWARD void setsym P((char *name, bin_off_t value));
 FORWARD void symres P((char *name));
-FORWARD void setseg P((unsigned newseg));
+FORWARD void setseg P((fastin_pt newseg));
 FORWARD void skip P((unsigned countsize));
 #ifdef EDOS
 FORWARD void writeheader P((char *commandname));
 #else
 FORWARD void writeheader P((void));
 #endif
-FORWARD void writenulls P((offset_t count));
-
-/* link all symbols connected to entry symbols */
-
-PUBLIC void linksyms(argreloc_output)
-bool_pt argreloc_output;
-{
-    char needlink;
-    struct entrylist *elptr;
-    struct modstruct *modptr;
-    struct symstruct *symptr;
-
-#ifdef BSD_A_OUT
-    reloc_output = argreloc_output;
-    if (argreloc_output)
-    {
-	if (modfirst->modnext != NULL)
-	    fatalerror("relocatable output only works for one input file");
-	for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
-	    modptr->loadflag = TRUE;
-	return;
-    }
-#endif
-    if ((symptr = findsym("_main")) != NULL)
-	entrysym(symptr);
-    do
-    {
-	if ((elptr = entryfirst) == NULL)
-	    fatalerror("no start symbol");
-	for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
-	    modptr->loadflag = FALSE;
-	for (; elptr != NULL; elptr = elptr->elnext)
-	    linkrefs(elptr->elsymptr->modptr);
-	if ((symptr = findsym("start")) != NULL ||
-	    (symptr = findsym("crtso")) != NULL)
-	    linkrefs(symptr->modptr);
-	needlink = FALSE;
-	{
-	    struct redlist *prlptr;
-	    struct redlist *rlptr;
-
-	    for (rlptr = redfirst; rlptr != NULL;
-		 rlptr = (prlptr = rlptr)->rlnext)
-		if (rlptr->rlmodptr->loadflag &&
-		    rlptr->rlmodptr->class > rlptr->rlsymptr->modptr->class)
-		{
-		    rlptr->rlsymptr->modptr = rlptr->rlmodptr;
-		    rlptr->rlsymptr->value = rlptr->rlvalue;
-		    if (rlptr == redfirst)
-			redfirst = rlptr->rlnext;
-		    else
-			prlptr->rlnext = rlptr->rlnext;
-		    needlink = TRUE;
-		}
-	}
-    }
-    while (needlink);
-}
+FORWARD void writenulls P((bin_off_t count));
 
 /* write binary file */
+#ifndef FUNCNAME 
+#define FUNCNAME writebin
+#endif
 
-PUBLIC void writebin(outfilename, argsepid, argbits32, argstripflag, arguzp)
+PUBLIC void FUNCNAME(outfilename, argsepid, argbits32, argstripflag, arguzp)
 char *outfilename;
 bool_pt argsepid;
 bool_pt argbits32;
@@ -201,15 +142,14 @@ bool_pt arguzp;
     struct nlist extsym;
     flags_t flags;
     struct modstruct *modptr;
-    char seg;
+    fastin_t seg;
     unsigned sizecount;
-    offset_t tempoffset;
-    struct symstruct *symptr;
+    bin_off_t tempoffset;
 
     sepid = argsepid;
     bits32 = argbits32;
     stripflag = argstripflag;
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
     uzp = arguzp && !reloc_output;
 #else
     uzp = arguzp;
@@ -217,15 +157,19 @@ bool_pt arguzp;
     if (uzp)
     {
 	if (btextoffset == 0)
+#ifdef QMAGIC
+	    btextoffset = page_size()+HEADERLEN;
+#else
 	    btextoffset = page_size();
+#endif
 	if (bdataoffset == 0 && sepid)
 	    bdataoffset = page_size();
     }
 #ifdef EDOS
     commandname = stralloc(outfilename);
-    if ((cptr = strchr(commandname, ':')) != NULL)
+    if ((cptr = strchr(commandname, ':')) != NUL_PTR)
 	commandname = cptr + 1;
-    if ((cptr = strrchr(commandname, '.')) != NULL)
+    if ((cptr = strrchr(commandname, '.')) != NUL_PTR)
 	*cptr = 0;
 #endif
 
@@ -241,6 +185,14 @@ bool_pt arguzp;
 	symres(segboundary);	/* __segXCL */
 	segboundary[7] = 'H';
 	symres(segboundary);	/* __segXCH */
+#ifndef DATASEGS
+        if( curseg > 3 )
+	{
+	   segboundary[6] = 'S';
+	   segboundary[7] = 'O';
+	   symres(segboundary); /* __segXSO */
+        }
+#endif
     }
 #ifdef EDOS
     curseg = 0;			/* data seg, s.b. variable */
@@ -251,18 +203,19 @@ bool_pt arguzp;
     symres("__end");
     curseg = 0;			/* text seg, s.b. variable */
     symres("__etext");
+    symres("__segoff");
 
     /* calculate segment and common sizes (sum over loaded modules) */
     /* use zero init of segsz[] */
     /* also relocate symbols relative to starts of their segments */
-    for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
+    for (modptr = modfirst; modptr != NUL_PTR; modptr = modptr->modnext)
 	if (modptr->loadflag)
 	{
 	    register struct symstruct **symparray;
 	    register struct symstruct *symptr;
 
 	    for (symparray = modptr->symparray;
-		 (symptr = *symparray) != NULL; ++symparray)
+		 (symptr = *symparray) != NUL_PTR; ++symparray)
 		if (symptr->modptr == modptr && !(symptr->flags & A_MASK))
 		{
 		    if (!(symptr->flags & (I_MASK | SA_MASK)))
@@ -274,14 +227,14 @@ bool_pt arguzp;
 		    }
 		    else if (symptr->value == 0)
 		    {
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 			if (!reloc_output)
 #endif
 			    undefined(symptr->name);
 		    }
 		    else
 		    {
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 #if 0
 			if (!reloc_output)
 #else	
@@ -289,7 +242,7 @@ bool_pt arguzp;
 #endif			
 #endif
 			{
-			    tempoffset = ld_roundup(symptr->value, 4, offset_t);
+			    tempoffset = ld_roundup(symptr->value, 4, bin_off_t);
 			    /* temp kludge quad alignment for 386 */
 			    symptr->value = comsz[seg = symptr->flags & SEGM_MASK];
 			    comsz[seg] += tempoffset;
@@ -306,8 +259,8 @@ bool_pt arguzp;
 
 		/* adjust sizes to even to get quad boundaries */
 		/* this should be specifiable dynamically */
-		segsz[seg] = ld_roundup(segsz[seg], 4, offset_t);
-		comsz[seg] = ld_roundup(comsz[seg], 4, offset_t);
+		segsz[seg] = ld_roundup(segsz[seg], 4, bin_off_t);
+		comsz[seg] = ld_roundup(comsz[seg], 4, bin_off_t);
 #endif
 		cptr += sizecount;
 	    }
@@ -315,7 +268,13 @@ bool_pt arguzp;
 
     /* calculate seg positions now their sizes are known */
     /* temp use fixed order 0D 0C 1D 1C 2D 2C ... */
-    /* assume seg 0 is text and rest are data */
+    /*
+#ifdef DATASEGS
+     * Assume seg 0 is text and rest are data
+#else
+     * Assume seg 1..3 are data, Seg 0 is real text, seg 4+ are far text
+#endif
+     */
 #ifdef EDOS
     if (btextoffset == 0)
 	btextoffset = binheaderlength(commandname);
@@ -326,16 +285,38 @@ bool_pt arguzp;
     etextpadoff = etextoffset = combase[0] + comsz[0];
     if (sepid)
     {
-	etextpadoff = ld_roundup(etextoffset, 0x10, offset_t);
+	etextpadoff = ld_roundup(etextoffset, 0x10, bin_off_t);
 	segadj[1] += etextpadoff - bdataoffset;
     }
+#ifdef QMAGIC
+    else if (uzp && bdataoffset == 0)
+    {
+	bdataoffset = ld_roundup(etextoffset, page_size(), bin_off_t);
+	etextpadoff = ld_roundup(etextoffset, page_size(), bin_off_t);
+	segadj[1] += etextpadoff - bdataoffset;
+    }
+#endif
     else if (bdataoffset == 0)
 	bdataoffset = etextpadoff;
     segpos[1] = segbase[1] = edataoffset = bdataoffset;
     combase[1] = segbase[1] + segsz[1];
+#ifndef DATASEGS
+    for (seg = 4; seg < NSEG; ++seg)
+    {
+	segpos[seg] = segbase[seg] = 0;
+	combase[seg] = segbase[seg] + segsz[seg];
+	segadj[seg] = etextpadoff;
+
+	etextpadoff += ld_roundup(segsz[seg] + comsz[seg], 0x10, bin_off_t);
+	segadj[1]   += ld_roundup(segsz[seg] + comsz[seg], 0x10, bin_off_t);
+    }
+    for (seg = 2; seg < 4; ++seg)
+#else
     for (seg = 2; seg < NSEG; ++seg)
+#endif
     {
 	segpos[seg] = segbase[seg] = combase[seg - 1] + comsz[seg - 1];
+#ifdef MC6809
 	if (seg == DPSEG)
 	{
 	    /* temporarily have fixed DP seg */
@@ -343,27 +324,39 @@ bool_pt arguzp;
 	    tempoffset = segsz[seg] + comsz[seg];
 	    if (tempoffset > 0x100)
 		fatalerror("direct page segment too large");
-	    if ((((segbase[seg] + tempoffset) ^ segbase[seg]) & ~0xFF) != 0)
-		segpos[seg] = segbase[seg] = (segbase[seg] + 0xFF) & ~0xFF;
+	    if ((((segbase[seg] + tempoffset) ^ segbase[seg])
+		 & ~(bin_off_t) 0xFF) != 0)
+		segpos[seg] = segbase[seg] = (segbase[seg] + 0xFF)
+					     & ~(bin_off_t) 0xFF;
 	}
+#endif
+
+#ifdef QMAGIC
+	if(seg==3 && uzp && !stripflag) /* XXX Stripped last seek needed */
+	{
+	   bin_off_t val;
+	   val = ld_roundup(segbase[seg]+segsz[seg], page_size(), bin_off_t);
+	   segsz[seg] = val - segbase[seg];
+	}
+#endif 
 	combase[seg] = segbase[seg] + segsz[seg];
 	segadj[seg] = segadj[seg - 1];
     }
 
     /* relocate symbols by offsets of segments in memory */
-    for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
+    for (modptr = modfirst; modptr != NUL_PTR; modptr = modptr->modnext)
 	if (modptr->loadflag)
 	{
 	    register struct symstruct **symparray;
 	    register struct symstruct *symptr;
 
 	    for (symparray = modptr->symparray;
-		 (symptr = *symparray) != NULL; ++symparray)
+		 (symptr = *symparray) != NUL_PTR; ++symparray)
 		if (symptr->modptr == modptr && !(symptr->flags & A_MASK))
 		{
 		    if (symptr->flags & (C_MASK | SA_MASK))
 		    {
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 #if 0
 			if (!reloc_output)
 #else	
@@ -373,7 +366,7 @@ bool_pt arguzp;
 			    symptr->value += combase[symptr->flags & SEGM_MASK];
 		    }
 		    else
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 		    if (!reloc_output || !(symptr->flags & I_MASK))
 #endif
 			symptr->value += segbase[symptr->flags & SEGM_MASK];
@@ -383,8 +376,12 @@ bool_pt arguzp;
     /* adjust special symbols */
     for (seg = 0; seg < NSEG; ++seg)
     {
+#ifdef DATASEGS
 	if (segsz[seg] != 0)
 	    /* only count data of nonzero length */
+#else
+	if (segsz[seg] != 0 && seg < 4)
+#endif
 	    edataoffset = segbase[seg] + segsz[seg];
 	segboundary[5] = hexdigit[seg];		/* to __segX?H */
 	segboundary[6] = 'D';
@@ -398,24 +395,45 @@ bool_pt arguzp;
 	segboundary[7] = 'H';
 	setsym(segboundary, tempoffset + comsz[seg]);
 						/* __segXCH */
+#ifndef DATASEGS
+        if( seg > 3 )
+	{
+	   segboundary[6] = 'S';
+	   segboundary[7] = 'O';
+	   setsym(segboundary, (bin_off_t)(segadj[seg]-segadj[0])/0x10);
+	   /* __segXSO */
+        }
+#endif
     }
     setsym("__etext", etextoffset);
     setsym("__edata", edataoffset);
+#ifdef DATASEGS
     setsym("__end", endoffset = combase[NSEG - 1] + comsz[NSEG - 1]);
+#else
+    setsym("__end", endoffset = combase[3] + comsz[3]);
+#endif
+    setsym("__segoff", (bin_off_t)(segadj[1]-segadj[0])/0x10);
+    if( !bits32 )
+    {
+        if( etextoffset > 65536L )
+            fatalerror("text segment too large for 16bit");
+        if( endoffset > 65536L )
+            fatalerror("data segment too large for 16bit");
+    }
 
     openout(outfilename);
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
     if (reloc_output)
 	seektrel(FILEHEADERLENGTH
-		 + (long) (etextpadoff - btextoffset)
-		 + (long) (edataoffset - bdataoffset));
+		 + (unsigned long) (etextpadoff - btextoffset)
+		 + (unsigned long) (edataoffset - bdataoffset));
 #endif
 #ifdef EDOS
     writeheader(commandname);
 #else
     writeheader();
 #endif
-    for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
+    for (modptr = modfirst; modptr != NUL_PTR; modptr = modptr->modnext)
 	if (modptr->loadflag)
 	{
 	    linkmod(modptr);
@@ -431,33 +449,44 @@ bool_pt arguzp;
 #endif
 
 	seekout(FILEHEADERLENGTH
-		+ (long) (etextpadoff - btextoffset)
-		+ (long) (edataoffset - bdataoffset)
-#ifdef BSD_A_OUT
-		+ ((long) ndreloc + ntreloc)
-		  * (long) sizeof(struct relocation_info)
+		+ (unsigned long) (etextpadoff - btextoffset)
+		+ (unsigned long) (edataoffset - bdataoffset)
+#ifdef REL_OUTPUT
+		+ ((unsigned long) ndreloc + ntreloc) * RELOC_INFO_SIZE
 #endif
 		);
 	extsym.n_was_numaux = extsym.n_was_type = 0;
 #ifdef BSD_A_OUT
 	stringoff = 4;
 #endif
-	for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
+	for (modptr = modfirst; modptr != NUL_PTR; modptr = modptr->modnext)
 	    if (modptr->loadflag)
 	    {
 		register struct symstruct **symparray;
 		register struct symstruct *symptr;
 
 		for (symparray = modptr->symparray;
-		     (symptr = *symparray) != NULL; ++symparray)
+		     (symptr = *symparray) != NUL_PTR; ++symparray)
 		    if (symptr->modptr == modptr)
 		    {
 #ifdef BSD_A_OUT
 			offtocn((char *) &extsym.n_was_strx,
-				(offset_t) stringoff, 4);
+				(bin_off_t) stringoff, 4);
 #else
-			strncpy((char *) &extsym.n_was_name, symptr->name,
+#if ELF_SYMS
+			if( symptr->name[0] == '_' && symptr->name[1] )
+			   strncpy((char *) extsym.n_was_name, symptr->name+1,
+				   sizeof extsym.n_was_name);
+			else
+			{
+			   memcpy((char *) extsym.n_was_name, "__", 2);
+			   strncpy((char *) extsym.n_was_name+2, symptr->name,
+				   sizeof(extsym.n_was_name)-2);
+			}
+#else
+			strncpy((char *) extsym.n_was_name, symptr->name,
 				sizeof extsym.n_was_name);
+#endif
 #endif
 			u4cn((char *) &extsym.n_value, (u4_t) symptr->value,
 			     sizeof extsym.n_value);
@@ -467,18 +496,27 @@ bool_pt arguzp;
 			    extsym.n_was_sclass = C_EXT;
 			else
 			    extsym.n_was_sclass = C_STAT;
-			if (!(flags & I_MASK) ||
-#ifdef BSD_A_OUT
+			if (!(flags & I_MASK) || (
+#ifdef REL_OUTPUT
 			     !reloc_output &&
 #endif
-			     flags & C_MASK)
+			     (flags & C_MASK)))
 			    switch (flags & (A_MASK | SEGM_MASK))
 			    {
-			    case 0:
+#ifdef DATASEGS
+ 			    case 0:
+#else
+			    default:
+#endif
 				extsym.n_was_sclass |= N_TEXT;
 			    case A_MASK:
 				break;
-			    default:
+#ifdef DATASEGS
+ 			    default:
+#else
+			    case 1: case 2: case 3:
+			    case A_MASK|1: case A_MASK|2: case A_MASK|3:
+#endif
 				if (flags & (C_MASK | SA_MASK))
 				    extsym.n_was_sclass |= N_BSS;
 				else
@@ -488,43 +526,64 @@ bool_pt arguzp;
 			writeout((char *) &extsym, sizeof extsym);
 			++nsym;
 #ifdef BSD_A_OUT
+#if ELF_SYMS
+			stringoff += strlen(symptr->name);
+			if( symptr->name[0] != '_' || symptr->name[1] == '\0' )
+			   stringoff += 3;
+#else
 			stringoff += strlen(symptr->name) + 1;
+#endif
 #endif
 		    }
 	    }
 #ifdef BSD_A_OUT
-	offtocn((char *) &extsym.n_was_strx, (offset_t) stringoff, 4);
+	offtocn((char *) &extsym.n_was_strx, (bin_off_t) stringoff, 4);
 	writeout((char *) &extsym.n_was_strx, 4);
-	for (modptr = modfirst; modptr != NULL; modptr = modptr->modnext)
+	for (modptr = modfirst; modptr != NUL_PTR; modptr = modptr->modnext)
 	    if (modptr->loadflag)
 	    {
 		register struct symstruct **symparray;
 		register struct symstruct *symptr;
 
 		for (symparray = modptr->symparray;
-		     (symptr = *symparray) != NULL; ++symparray)
+		     (symptr = *symparray) != NUL_PTR; ++symparray)
 		    if (symptr->modptr == modptr)
+#if ELF_SYMS
+                    {
+		       if( symptr->name[0] == '_' && symptr->name[1] )
+			  writeout(symptr->name + 1, strlen(symptr->name));
+		       else
+		       {
+		          writeout("__", 2);
+			  writeout(symptr->name, strlen(symptr->name) + 1);
+		       }
+		    }
+#else
 			writeout(symptr->name, strlen(symptr->name) + 1);
+#endif
 	    }
 #endif
-	seekout((long) offsetof(struct exec, a_syms));
+	seekout((unsigned long) offsetof(struct exec, a_syms));
 	u4cn(buf4, (u4_t) nsym * sizeof extsym,
 	     memsizeof(struct exec, a_syms));
 	writeout(buf4, memsizeof(struct exec, a_syms));
-#if defined (BSD_A_OUT) && !defined(__linux__)
-	seekout((long) offsetof(struct exec, a_trsize));
-	u4cn(buf4, (u4_t) ntreloc * sizeof(struct relocation_info),
-	     memsizeof(struct exec, a_trsize));
-	writeout(buf4, memsizeof(struct exec, a_trsize));
-	seekout((long) offsetof(struct exec, a_drsize));
-	u4cn(buf4, (u4_t) ndreloc * sizeof(struct relocation_info),
-	     memsizeof(struct exec, a_drsize));
-	writeout(buf4, memsizeof(struct exec, a_drsize));
+#ifdef REL_OUTPUT
+	if( FILEHEADERLENGTH >= offsetof(struct exec, a_trsize)+8)
+	{
+		seekout((unsigned long) offsetof(struct exec, a_trsize));
+		u4cn(buf4, (u4_t) ntreloc * RELOC_INFO_SIZE,
+	     	memsizeof(struct exec, a_trsize));
+		writeout(buf4, memsizeof(struct exec, a_trsize));
+		seekout((unsigned long) offsetof(struct exec, a_drsize));
+		u4cn(buf4, (u4_t) ndreloc * RELOC_INFO_SIZE,
+	     	memsizeof(struct exec, a_drsize));
+		writeout(buf4, memsizeof(struct exec, a_drsize));
+	}
 #endif
     }
 #endif /* MINIX */
     closeout();
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
     if (!reloc_output)
 #endif
 	executable();
@@ -542,12 +601,12 @@ char *commandname;
 
     count = 2 + 2 + 1;		/* len len nul */
     startptr = findsym("start");
-    for (elptr = entryfirst; elptr != NULL; elptr = elptr->elnext)
+    for (elptr = entryfirst; elptr != NUL_PTR; elptr = elptr->elnext)
     {
 	name = idconvert(elptr, commandname);
 	count += strlen(name) + 1 + 2 + 1;	/* nul off flg */
 	ourfree(name);
-	if (startptr != NULL)
+	if (startptr != NUL_PTR)
 	    count += 6;		/* LBSR $xxxx and LBRA $xxxx */
     }
     return count;
@@ -597,7 +656,7 @@ struct modstruct *modptr;
     char buf[ABS_TEXT_MAX];
     int command;
     unsigned char modify;
-    offset_t offset;
+    bin_off_t offset;
     int symbolnum;
     struct symstruct **symparray;
     struct symstruct *symptr;
@@ -606,7 +665,7 @@ struct modstruct *modptr;
     relocsize = 2;
     symparray = modptr->symparray;
     openin(modptr->filename);	/* does nothing if already open */
-    seekin((long) modptr->textoffset);
+    seekin(modptr->textoffset);
     while (TRUE)
     {
 	if ((command = readchar()) < 0)
@@ -645,7 +704,7 @@ struct modstruct *modptr;
 	    default:
 		if ((modify -= CM_0_SEG) >= NSEG)
 		    inputerror("bad data in");
-		setseg((unsigned) modify);
+		setseg(modify);
 		break;
 	    }
 	    break;
@@ -659,10 +718,17 @@ struct modstruct *modptr;
 	case CM_OFFSET_RELOC:
 	    offset = readsize(relocsize);
 	    if (modify & R_MASK)
+	    {
+#ifndef DATASEGS
+                int m = (modify & SEGM_MASK);
+	        if( curseg != m && m != SEGM_MASK )
+	           interseg(modptr->filename, modptr->archentry, (char*)0);
+#endif
 		offset -= (spos + relocsize);
+            }
 	    offtocn(buf, segbase[modify & SEGM_MASK] + offset, relocsize);
 	    writeout(buf, relocsize);
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 	    if (reloc_output)
 	    {
 		u4_t bitfield;
@@ -695,7 +761,7 @@ struct modstruct *modptr;
 		else
 		    writedrel(buf, 4);
 	    }
-#endif /* BSD_A_OUT */
+#endif /* REL_OUTPUT */
 	    spos += relocsize;
 	    break;
 	case CM_SYMBOL_RELOC:
@@ -703,14 +769,21 @@ struct modstruct *modptr;
 					    (modify & S_MASK ? 2 : 1))];
 	    offset = readconvsize((unsigned) modify & OF_MASK);
 	    if (modify & R_MASK)
+	    {
+#ifndef DATASEGS
+                int m = (symptr->flags & SEGM_MASK);
+	        if( curseg != m && m != SEGM_MASK )
+	           interseg(modptr->filename, modptr->archentry, symptr->name);
+#endif
 		offset -= (spos + relocsize);
-#ifdef BSD_A_OUT
+	    }
+#ifdef REL_OUTPUT
 	    if (!reloc_output || !(symptr->flags & I_MASK))
 #endif
 		offset += symptr->value;	    
 	    offtocn(buf, offset, relocsize);
 	    writeout(buf, relocsize);
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 	    if (reloc_output)
 	    {
 		u4_t bitfield;
@@ -747,31 +820,18 @@ struct modstruct *modptr;
 		else
 		    writedrel(buf, 4);
 	    }
-#endif /* BSD_A_OUT */
+#endif /* REL_OUTPUT */
 	    spos += relocsize;
 	}
     }
 }
 
-PRIVATE void linkrefs(modptr)
-struct modstruct *modptr;
-{
-    register struct symstruct **symparray;
-    register struct symstruct *symptr;
-
-    modptr->loadflag = TRUE;
-    for (symparray = modptr->symparray;
-	 (symptr = *symparray) != NULL; ++symparray)
-	if (symptr->modptr->loadflag == FALSE)
-	    linkrefs(symptr->modptr);
-}
-
 PRIVATE void padmod(modptr)
 struct modstruct *modptr;
 {
-    offset_t count;
-    char seg;
-    offset_t size;
+    bin_off_t count;
+    fastin_t seg;
+    bin_off_t size;
     unsigned sizecount;
     char *sizeptr;
 
@@ -785,7 +845,7 @@ struct modstruct *modptr;
 
 	/* pad to quad boundary */
 	/* not padding in-between common areas which sometimes get into file */
-	if ((size = ld_roundup(segpos[seg], 4, offset_t) - segpos[seg]) != 0)
+	if ((size = ld_roundup(segpos[seg], 4, bin_off_t) - segpos[seg]) != 0)
 	{
 	    setseg(seg);
 	    writenulls(size);
@@ -797,14 +857,14 @@ struct modstruct *modptr;
 
 PRIVATE void setsym(name, value)
 char *name;
-offset_t value;
+bin_off_t value;
 {
     struct symstruct *symptr;
 
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
     if (!reloc_output)
 #endif
-	if ((symptr = findsym(name)) != NULL)
+	if ((symptr = findsym(name)) != NUL_PTR)
 	    symptr->value = value;
 }
 
@@ -813,13 +873,13 @@ register char *name;
 {
     register struct symstruct *symptr;
 
-    if ((symptr = findsym(name)) != NULL)
+    if ((symptr = findsym(name)) != NUL_PTR)
     {
 	if ((symptr->flags & SEGM_MASK) == SEGM_MASK)
 	    symptr->flags &= ~SEGM_MASK | curseg;
 	if (symptr->flags != (I_MASK | curseg) || symptr->value != 0)
 	    reserved(name);
-#ifdef BSD_A_OUT
+#ifdef REL_OUTPUT
 	if (!reloc_output)
 #endif
 	    symptr->flags = E_MASK | curseg;	/* show defined, not common */
@@ -829,20 +889,21 @@ register char *name;
 /* set new segment */
 
 PRIVATE void setseg(newseg)
-unsigned newseg;
+fastin_pt newseg;
 {
     if (newseg != curseg)
     {
 	segpos[curseg] = spos;
 	spos = segpos[curseg = newseg];
-	seekout(FILEHEADERLENGTH + (long) spos + (long) segadj[curseg]);
+	seekout(FILEHEADERLENGTH + (unsigned long) spos
+		+ (unsigned long) segadj[curseg]);
     }
 }
 
 PRIVATE void skip(countsize)
 unsigned countsize;
 {
-    writenulls((offset_t) readsize(countsize));
+    writenulls((bin_off_t) readsize(countsize));
 }
 
 #ifdef EDOS
@@ -851,14 +912,14 @@ PRIVATE void writeheader(commandname)
 char *commandname;
 {
     char buf[MAX_OFFSET_SIZE];
-    offset_t offset;
+    bin_off_t offset;
     unsigned headlength;
     char *name;
     struct entrylist *elptr;
     struct symstruct *startptr;
 
     headlength = binheaderlength(commandname);
-    for (elptr = entryfirst; elptr != NULL; elptr = elptr->elnext)
+    for (elptr = entryfirst; elptr != NUL_PTR; elptr = elptr->elnext)
 	headlength -= 6;
     offset = headlength;
     startptr = findsym("start");
@@ -866,21 +927,21 @@ char *commandname;
     writeout(buf, 2);
     writechar(0xFF);		/* dummy data length 0xFFFF takes everything */
     writechar(0xFF);
-    for (elptr = entryfirst; elptr != NULL; elptr = elptr->elnext)
+    for (elptr = entryfirst; elptr != NUL_PTR; elptr = elptr->elnext)
     {
 	name = idconvert(elptr, commandname);
 	writeout(name, (unsigned) strlen(name) + 1);
 	ourfree(name);
-	offtocn(buf, startptr == NULL ? elptr->elsymptr->value : offset, 2);
+	offtocn(buf, startptr == NUL_PTR ? elptr->elsymptr->value : offset, 2);
 	writeout(buf, 2);
 	writechar(0x82);	/* 8 = set flags from here, 2 = cmd line */
 	offset += 6;		/* LBSR $xxxx and LBRA $xxxx */
     }
     writechar(0);
-    if (startptr != NULL)
+    if (startptr != NUL_PTR)
     {
 	offset = headlength + 3;	/* over 1st LBSR */
-	for (elptr = entryfirst; elptr != NULL; elptr = elptr->elnext)
+	for (elptr = entryfirst; elptr != NUL_PTR; elptr = elptr->elnext)
 	{
 	    writechar(0x17);	/* LBSR */
 	    offtocn(buf, startptr->value - offset, 2);
@@ -902,48 +963,109 @@ PRIVATE void writeheader()
     struct exec header;
 
     memset(&header, 0, sizeof header);
+#ifdef STANDARD_GNU_A_OUT
+#ifdef N_SET_MAGIC
+#ifdef QMAGIC
+    if(uzp)
+       N_SET_MAGIC(header, QMAGIC);
+    else
+#endif
+    N_SET_MAGIC(header, OMAGIC);
+#else
+    *(unsigned short *) &header.a_magic = OMAGIC;  /* XXX - works for 386BSD */
+#endif
+#else
     header.a_magic[0] = A_MAGIC0;
     header.a_magic[1] = A_MAGIC1;
-#ifdef BSD_A_OUT
+#endif
+#ifdef REL_OUTPUT
     if (!reloc_output)
 #endif
     {
+#ifdef STANDARD_GNU_A_OUT
+#ifdef N_SET_FLAGS
+	N_SET_FLAGS(header, 0);
+#else
+	/* XXX - works for 386BSD */
+#endif
+#else
 	header.a_flags = sepid ? A_SEP : A_EXEC;
 	if (uzp)
 	    header.a_flags |= A_UZP;
+#endif
     }
 #ifdef BSD_A_OUT
+#ifdef STANDARD_GNU_A_OUT
+#ifdef N_SET_FLAGS
+    N_SET_MACHTYPE(header, M_386);
+#else
+	/* XXX - works for 386BSD which doesn't define its own machtype :-( */
+#endif
+#else
     header.a_cpu = (bits32 || reloc_output) ? A_I80386 : A_I8086;
+#endif
 #else
     header.a_cpu = bits32 ? A_I80386 : A_I8086;
 #endif
+#ifndef STANDARD_GNU_A_OUT
     header.a_hdrlen = FILEHEADERLENGTH;
+#endif
+#ifdef QMAGIC
+    if (uzp)
+       offtocn((char *) &header.a_text, etextpadoff - btextoffset+HEADERLEN,
+	       sizeof header.a_text);
+    else
+#endif
     offtocn((char *) &header.a_text, etextpadoff - btextoffset,
 	    sizeof header.a_text);
     offtocn((char *) &header.a_data, edataoffset - bdataoffset,
 	    sizeof header.a_data);
     offtocn((char *) &header.a_bss, endoffset - edataoffset,
 	    sizeof header.a_bss);
-#ifdef BSD_A_OUT
+
+#ifdef REL_OUTPUT
     if (!reloc_output)
 #endif
     {
-	if (uzp)
-	    offtocn((char *) &header.a_entry, page_size(),
+	offtocn((char *) &header.a_entry, btextoffset,
 		    sizeof header.a_entry);
-	offtocn((char *) &header.a_total, (offset_t)
+#ifndef STANDARD_GNU_A_OUT
+	offtocn((char *) &header.a_total, (bin_off_t)
     	    (endoffset < 0x00010000L ? 0x00010000L : endoffset + 0x0008000L),
 		sizeof header.a_total);
+#endif
     }
-    writeout((char *) &header, FILEHEADERLENGTH);
+    if( FILEHEADERLENGTH )
+       writeout((char *) &header, FILEHEADERLENGTH);
 }
 
 #endif /* MINIX */
 
 PRIVATE void writenulls(count)
-offset_t count;
+bin_off_t count;
 {
+    long lcount = count;
+    if( lcount < 0 )
+    	fatalerror("org command requires reverse seek");
     spos += count;
-    while (count--)
+    while (count-- > 0)
 	writechar(0);
 }
+#else
+
+#ifndef FUNCNAME
+#define FUNCNAME writebin
+#endif
+
+PUBLIC void FUNCNAME(outfilename, argsepid, argbits32, argstripflag, arguzp)
+char *outfilename;
+bool_pt argsepid;
+bool_pt argbits32;
+bool_pt argstripflag;
+bool_pt arguzp;
+{
+    char * s  = "WARNING: Native a.out generation not included, sorry\n";
+    write(2, s, strlen(s));
+    write_elks(outfilename, argsepid, argbits32, argstripflag, arguzp, 0);
+}
+#endif
